from __future__ import annotations

import ast
from typing import cast

import pytest

from flake8_pydantic._utils import is_pydantic_model

# Positive cases:
SUBCLASSES_BASE_MODEL_1 = """
class Model(BaseModel):
    pass
"""

SUBCLASSES_BASE_MODEL_2 = """
class Model(pydantic.BaseModel):
    pass
"""

SUBCLASSES_ROOT_MODEL = """
class Model(RootModel):
    root: int
"""

HAS_ANNOTATED_MODEL_CONFIG = """
class SubModel(ParentModel):
    model_config: ModelConfig = {}
"""

HAS_MODEL_CONFIG = """
class SubModel(ParentModel):
    model_config = {}
"""

HAS_FIELD_FUNCTION_1 = """
class SubModel(ParentModel):
    a = Field()
"""

HAS_FIELD_FUNCTION_2 = """
class SubModel(ParentModel):
    a: int = Field()
"""

HAS_FIELD_FUNCTION_3 = """
class SubModel(ParentModel):
    a = pydantic.Field()
"""

HAS_FIELD_FUNCTION_4 = """
class SubModel(ParentModel):
    a: int = pydantic.Field()
"""

USES_ANNOTATED_1 = """
class SubModel(ParentModel):
    a: Annotated[int, ""]
"""

USES_ANNOTATED_2 = """
class SubModel(ParentModel):
    a: typing.Annotated[int, ""]
"""

HAS_PYDANTIC_DECORATOR_1 = """
class SubModel(ParentModel):
    @computed_field
    @unrelated
    def func(): pass
"""

HAS_PYDANTIC_DECORATOR_2 = """
class SubModel(ParentModel):
    @pydantic.computed_field
    def func(): pass
"""

HAS_PYDANTIC_METHOD_1 = """
class SubModel(ParentModel):
    def model_dump(self): pass
"""

HAS_PYDANTIC_METHOD_2 = """
class SubModel(ParentModel):
    def __pydantic_some_method__(self): pass
"""

# Negative cases:
NO_BASES = """
class Model:
    a = Field()
"""


@pytest.mark.parametrize(
    ["source", "expected"],
    [
        (SUBCLASSES_BASE_MODEL_1, True),
        (SUBCLASSES_BASE_MODEL_2, True),
        (SUBCLASSES_ROOT_MODEL, True),
        (HAS_ANNOTATED_MODEL_CONFIG, True),
        (HAS_MODEL_CONFIG, True),
        (HAS_FIELD_FUNCTION_1, True),
        (HAS_FIELD_FUNCTION_2, True),
        (HAS_FIELD_FUNCTION_3, True),
        (HAS_FIELD_FUNCTION_4, True),
        (USES_ANNOTATED_1, True),
        (USES_ANNOTATED_2, True),
        (HAS_PYDANTIC_DECORATOR_1, True),
        (HAS_PYDANTIC_DECORATOR_2, True),
        (HAS_PYDANTIC_METHOD_1, True),
        (HAS_PYDANTIC_METHOD_2, True),
        (NO_BASES, False),
    ],
)
def test_is_pydantic_model(source: str, expected: bool) -> None:
    class_def = cast(ast.ClassDef, ast.parse(source).body[0])
    assert is_pydantic_model(class_def) == expected



from __future__ import annotations

import ast
from collections import deque
from typing import Literal

from ._compat import TypeAlias
from ._utils import extract_annotations, is_dataclass, is_function, is_name, is_pydantic_model
from .errors import PYD001, PYD002, PYD003, PYD004, PYD005, PYD006, PYD010, Error

ClassType: TypeAlias = Literal["pydantic_model", "dataclass", "other_class"]


class Visitor(ast.NodeVisitor):
    def __init__(self) -> None:
        self.errors: list[Error] = []
        self.class_stack: deque[ClassType] = deque()

    def enter_class(self, node: ast.ClassDef) -> None:
        if is_pydantic_model(node):
            self.class_stack.append("pydantic_model")
        elif is_dataclass(node):
            self.class_stack.append("dataclass")
        else:
            self.class_stack.append("other_class")

    def leave_class(self) -> None:
        self.class_stack.pop()

    @property
    def current_class(self) -> ClassType:
        return self.class_stack[-1]

    def _check_pyd_001(self, node: ast.AnnAssign) -> None:
        if (
            self.current_class in {"pydantic_model", "dataclass"}
            and isinstance(node.value, ast.Call)
            and is_function(node.value, "Field")
            and len(node.value.args) >= 1
        ):
            self.errors.append(PYD001.from_node(node))

    def _check_pyd_002(self, node: ast.ClassDef) -> None:
        if self.current_class == "pydantic_model":
            invalid_assignments = [
                assign
                for assign in node.body
                if isinstance(assign, ast.Assign)
                if isinstance(assign.targets[0], ast.Name)
                if not assign.targets[0].id.startswith("_")
            ]
            for assignment in invalid_assignments:
                self.errors.append(PYD002.from_node(assignment))

    def _check_pyd_003(self, node: ast.AnnAssign) -> None:
        if (
            self.current_class in {"pydantic_model", "dataclass"}
            and isinstance(node.value, ast.Call)
            and is_function(node.value, "Field")
            and len(node.value.keywords) == 1
            and node.value.keywords[0].arg == "default"
        ):
            self.errors.append(PYD003.from_node(node))

    def _check_pyd_004(self, node: ast.AnnAssign) -> None:
        if (
            self.current_class in {"pydantic_model", "dataclass"}
            and isinstance(node.annotation, ast.Subscript)
            and is_name(node.annotation.value, "Annotated")
            and isinstance(node.annotation.slice, ast.Tuple)
        ):
            field_call = next(
                (
                    elt
                    for elt in node.annotation.slice.elts
                    if isinstance(elt, ast.Call)
                    and is_function(elt, "Field")
                    and any(k.arg == "default" for k in elt.keywords)
                ),
                None,
            )
            if field_call is not None:
                self.errors.append(PYD004.from_node(node))

    def _check_pyd_005(self, node: ast.ClassDef) -> None:
        if self.current_class in {"pydantic_model", "dataclass"}:
            previous_targets: set[str] = set()

            for stmt in node.body:
                if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
                    # TODO only add before if AnnAssign?
                    # the following seems to work:
                    # date: date
                    previous_targets.add(stmt.target.id)
                    if previous_targets & extract_annotations(stmt.annotation):
                        self.errors.append(PYD005.from_node(stmt))

    def _check_pyd_006(self, node: ast.ClassDef) -> None:
        if self.current_class in {"pydantic_model", "dataclass"}:
            previous_targets: set[str] = set()

            for stmt in node.body:
                if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
                    if stmt.target.id in previous_targets:
                        self.errors.append(PYD006.from_node(stmt))

                    previous_targets.add(stmt.target.id)

    def _check_pyd_010(self, node: ast.ClassDef) -> None:
        if self.current_class == "other_class":
            for stmt in node.body:
                if (
                    isinstance(stmt, ast.AnnAssign)
                    and isinstance(stmt.target, ast.Name)
                    and stmt.target.id == "__pydantic_config__"
                ):
                    # __pydantic_config__: ... = ...
                    self.errors.append(PYD010.from_node(stmt))
                if isinstance(stmt, ast.Assign) and any(
                    t.id == "__pydantic_config__" for t in stmt.targets if isinstance(t, ast.Name)
                ):
                    # __pydantic_config__ = ...
                    self.errors.append(PYD010.from_node(stmt))

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.enter_class(node)
        self._check_pyd_002(node)
        self._check_pyd_005(node)
        self._check_pyd_006(node)
        self._check_pyd_010(node)
        self.generic_visit(node)
        self.leave_class()

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        self._check_pyd_001(node)
        self._check_pyd_003(node)
        self._check_pyd_004(node)
        self.generic_visit(node)



from __future__ import annotations

import ast


def get_decorator_names(decorator_list: list[ast.expr]) -> set[str]:
    names: set[str] = set()
    for dec in decorator_list:
        if isinstance(dec, ast.Call):
            names.add(dec.func.attr if isinstance(dec.func, ast.Attribute) else dec.func.id)  # type: ignore
        elif isinstance(dec, ast.Name):
            names.add(dec.id)
        elif isinstance(dec, ast.Attribute):
            names.add(dec.attr)

    return names


def _has_pydantic_model_base(node: ast.ClassDef, include_root_model: bool) -> bool:
    model_class_names = {"BaseModel"}
    if include_root_model:
        model_class_names.add("RootModel")

    for base in node.bases:
        if isinstance(base, ast.Name) and base.id in model_class_names:
            return True
        if isinstance(base, ast.Attribute) and base.attr in model_class_names:
            return True
    return False


def _has_model_config(node: ast.ClassDef) -> bool:
    for stmt in node.body:
        if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name) and stmt.target.id == "model_config":
            # model_config: ... = ...
            return True
        if isinstance(stmt, ast.Assign) and any(
            t.id == "model_config" for t in stmt.targets if isinstance(t, ast.Name)
        ):
            # model_config = ...
            return True
    return False


def _has_field_function(node: ast.ClassDef) -> bool:
    for stmt in node.body:
        if isinstance(stmt, (ast.Assign, ast.AnnAssign)) and isinstance(stmt.value, ast.Call):
            if isinstance(stmt.value.func, ast.Name) and stmt.value.func.id == "Field":
                # f = Field(...)
                return True
            if isinstance(stmt.value.func, ast.Attribute) and stmt.value.func.attr == "Field":
                # f = pydantic.Field(...)
                return True
    return False


def _has_annotated_field(node: ast.ClassDef) -> bool:
    for stmt in node.body:
        if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.annotation, ast.Subscript):
            if isinstance(stmt.annotation.value, ast.Name) and stmt.annotation.value.id == "Annotated":
                # f: Annotated[...]
                return True
            if isinstance(stmt.annotation.value, ast.Attribute) and stmt.annotation.value.attr == "Annotated":
                # f: typing.Annotated[...]
                return True
    return False


PYDANTIC_DECORATORS = {
    "computed_field",
    "field_serializer",
    "model_serializer",
    "field_validator",
    "model_validator",
}


def _has_pydantic_decorator(node: ast.ClassDef) -> bool:
    for stmt in node.body:
        if isinstance(stmt, ast.FunctionDef):
            decorator_names = get_decorator_names(stmt.decorator_list)
            if PYDANTIC_DECORATORS & decorator_names:
                return True
    return False


def _has_pydantic_method(node: ast.ClassDef) -> bool:
    for stmt in node.body:
        if isinstance(stmt, ast.FunctionDef) and stmt.name.startswith(("model_", "__pydantic_")):
            return True
    return False


def is_pydantic_model(node: ast.ClassDef, include_root_model: bool = True) -> bool:
    """Determine if a class definition is a Pydantic model.

    Multiple heuristics are use to determine if this is the case:
    - The class inherits from `BaseModel` (or `RootModel` if `include_root_model` is `True`).
    - The class has a `model_config` attribute set.
    - The class has a field defined with the `Field` function.
    - The class has a field making use of `Annotated`.
    - The class makes use of Pydantic decorators, such as `computed_field` or `model_validator`.
    - The class overrides any of the Pydantic methods, such as `model_dump`.
    """
    if not node.bases:
        return False

    return (
        _has_pydantic_model_base(node, include_root_model)
        or _has_model_config(node)
        or _has_field_function(node)
        or _has_annotated_field(node)
        or _has_pydantic_decorator(node)
        or _has_pydantic_method(node)
    )


def is_dataclass(node: ast.ClassDef) -> bool:
    """Determine if a class is a dataclass."""

    return bool({"dataclass", "pydantic_dataclass"} & get_decorator_names(node.decorator_list))


def is_function(node: ast.Call, function_name: str) -> bool:
    return (
        isinstance(node.func, ast.Name)
        and node.func.id == function_name
        or isinstance(node.func, ast.Attribute)
        and node.func.attr == function_name
    )


def is_name(node: ast.expr, name: str) -> bool:
    return isinstance(node, ast.Name) and node.id == name or isinstance(node, ast.Attribute) and node.attr == name


def extract_annotations(node: ast.expr) -> set[str]:
    annotations: set[str] = set()

    if isinstance(node, ast.Name):
        # foo: date = ...
        annotations.add(node.id)
    if isinstance(node, ast.BinOp):
        # foo: date | None = ...
        annotations |= extract_annotations(node.left)
        annotations |= extract_annotations(node.right)
    if isinstance(node, ast.Subscript):
        # foo: dict[str, date]
        # foo: Annotated[list[date], ...]
        if isinstance(node.slice, ast.Tuple):
            for elt in node.slice.elts:
                annotations |= extract_annotations(elt)
        if isinstance(node.slice, ast.Name):
            annotations.add(node.slice.id)

    return annotations



from __future__ import annotations

import ast
from collections.abc import Iterator
from importlib.metadata import version
from typing import Any

from .visitor import Visitor


class Plugin:
    name = "flake8-pydantic"
    version = version(name)

    def __init__(self, tree: ast.AST) -> None:
        self._tree = tree

    def run(self) -> Iterator[tuple[int, int, str, type[Any]]]:
        visitor = Visitor()
        visitor.visit(self._tree)
        for error in visitor.errors:
            yield *error.as_flake8_error(), type(self)



from __future__ import annotations

import ast
from abc import ABC
from dataclasses import dataclass
from typing import ClassVar

from ._compat import Self


@dataclass
class Error(ABC):
    error_code: ClassVar[str]
    message: ClassVar[str]
    lineno: int
    col_offset: int

    @classmethod
    def from_node(cls, node: ast.AST) -> Self:
        return cls(lineno=node.lineno, col_offset=node.col_offset)

    def as_flake8_error(self) -> tuple[int, int, str]:
        return (self.lineno, self.col_offset, f"{self.error_code} {self.message}")


class PYD001(Error):
    error_code = "PYD001"
    message = "Positional argument for Field default argument"


class PYD002(Error):
    error_code = "PYD002"
    message = "Non-annotated attribute inside Pydantic model"


class PYD003(Error):
    error_code = "PYD003"
    message = "Unecessary Field call to specify a default value"


class PYD004(Error):
    error_code = "PYD004"
    message = "Default argument specified in annotated"


class PYD005(Error):
    error_code = "PYD005"
    message = "Field name overrides annotation"


class PYD006(Error):
    error_code = "PYD006"
    message = "Duplicate field name"


class PYD010(Error):
    error_code = "PYD010"
    message = "Usage of __pydantic_config__"



from __future__ import annotations

import ast
from typing import cast

import pytest

from flake8_pydantic._utils import is_dataclass

# Positive cases:
DATACLASS_1 = """
@dataclass
class Model:
    pass
"""

DATACLASS_2 = """
@pydantic_dataclass
class Model:
    pass
"""

DATACLASS_3 = """
@dataclasses.dataclass
class Model:
    pass
"""

DATACLASS_4 = """
@dataclasses.dataclass()
@otherdec(arg=1)
class Model:
    pass
"""


@pytest.mark.parametrize(
    ["source", "expected"],
    [
        (DATACLASS_1, True),
        (DATACLASS_2, True),
        (DATACLASS_3, True),
        (DATACLASS_4, True),
    ],
)
def test_is_dataclass(source: str, expected: bool) -> None:
    class_def = cast(ast.ClassDef, ast.parse(source).body[0])
    assert is_dataclass(class_def) == expected
